/*
 * This file is part of ascad.
 *
 * Ascad is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ascad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2005 Andrei Krüger */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h> //vfprintf, etc.
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>
#include <math.h>

#define ASCAD_MAIN
#include "ascad.h"
#include "ocad.h"
#include "md5.h"
#include "list.h"

static char *_version_ = "ascad 0.8.15";

char *dospath = "C:\\Temp";
dword AS_mtime;

void xperror(const char *fmt, ...) {
  va_list ap;

  if (!errstr) errstr = strerror(errno);

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, ": %s\n", errstr);
  va_end(ap);
  exit(1);
}

void xerror(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}

void *xmalloc(size_t size) {
  void *p = malloc(size);
  if (!p) xerror("Out of memory");
  return p;
}


void DEB(const char *fmt, ...) {
#ifdef DEBUGGING
  va_list ap; va_start(ap, fmt); vfprintf(stderr, fmt, ap); fflush(stderr); va_end(ap);
#endif
}

#define BUF_SIZE 8192
#define LINE_SIZE 1024
#define isdelim(x) (isspace(x) || (x)==',')

char * strdup2(char *s1, char *s2)
{
  char *res;
  int n = strlen(s1);

  res = xmalloc(n + strlen(s2) + 1);
  strcpy(res, s1);
  strcpy(res+n, s2);
  return res;
}

char *skip_str(char **str, char *sub) {
  char *a = *str;
  while(*a && isspace(*a)) a++;
  while(*a && *a == *sub) a++, sub++;
  if (*sub != 0) return NULL;
  return *str = a;
}

char *get_num(char **str, double *num, double scale) {
  char *end;
  *num = strtod(*str, &end) * scale;
  return end == *str ? NULL : (*str = end);
}

char *get_2num(char **str, double *num1, double *num2, double scale) {
  char *save = *str;
  if (get_num(str, num1, scale) && skip_str(str, ",") && get_num(str, num2, scale))
    return *str;
  *str = save;
  return NULL;
}

char *get_long(char **str, long *num) {
  char *end;
  *num = strtol(*str, &end, 10);
  return end == *str ? NULL : (*str = end);
}

char *get_3long(char **str, long *num1, long *num2, long *num3) {
  char *save = *str;
  if (get_long(str, num1) && skip_str(str, ",") &&
      get_long(str, num2) && skip_str(str, ",") &&
      get_long(str, num3))
    return *str;
  *str = save;
  return NULL;
}

char *get_word(char **str, char **new)
{
  char *a = *str, *a0;
  int len;
  while(*a && isspace(*a)) a++; a0 = a;
  while(*a && !isdelim(*a)) a++;
  if ((len = a-a0) == 0) return NULL;
  *new = xmalloc(len+1);
  strncpy(*new, a0, len);
  (*new)[len] = 0;
  return *str = a;
}

void fit_page(point *loc, rect *page) {
  /* use right/upper side as reference if negative location */
  if (loc->x < 0) loc->x += page->x2;
  if (loc->y < 0) loc->y += page->y2;
}

time_t read_mtime(char *fn)
{
  struct stat st;
  if (stat(fn, &st))
    xperror(fn);
  return st.st_mtime;
}

int linecmp(char *a, char *b) {
  while(*a && *a != '\n' && *a == *b) a++, b++;
  return ((*a && *a != '\n') || (*b && *b != '\n'));
}

source_t * find_source(char *line, struct list_head *source_list) {
  source_t *source;

  list_for_each_entry(source, source_list, list)
    if (linecmp(line, source->config) == 0)
      return source;
  return NULL;
}

int read_hash(FILE *fp, char *hash)
{
  char line[LINE_SIZE];

  memset(hash, '.', HASH_LEN);
  hash[HASH_LEN] = '\0';

  if (fseek(fp, 0, SEEK_SET))
    return -1;

  while (fgets(line, LINE_SIZE, fp))
  {
    if (line[0] != '%') break;
    if (strncmp(line, "%ASCAD HASH ", 12) == 0)
    {
      strncpy(hash, line+12, HASH_LEN);
      break;
    }
  }

  return 0;
}

void calc_map_hash(map_t *map)
{
  md5_context ctx;
  source_t *source;

  md5_starts( &ctx );
  md5_update( &ctx, (uint8 *)map->config, strlen(map->config));
  list_for_each_entry(source, &map->source_list, list)
    md5_update( &ctx, (uint8 *)source->config, strlen(source->config));
  md5_finish_hex( &ctx, map->hash );
}

int find_encodes(FILE *cache_fp, struct list_head *source_list, rect *page)
{
  /* Find encoded data blocks from cache or recreate them from sources */
  char line[LINE_SIZE];
  source_t *source;

  /* read existing blocks from cache */
  if (cache_fp != NULL)
  {
    if(fseek(cache_fp, 0, SEEK_SET))
      return -1;

    while (fgets(line, LINE_SIZE, cache_fp))
    {
      if (strncmp(line, "%ASCAD BEGIN ", 13) == 0)
      {
        if ((source = find_source(line+13, source_list)) != NULL)
        {
          source->encoded.fp = cache_fp;
          source->encoded.start = ftell(cache_fp);
          while (fgets(line, LINE_SIZE, cache_fp))
            if (strncmp(line, "%ASCAD END ", 11) == 0)
              break;
          source->encoded.length = ftell(cache_fp) - strlen(line) - source->encoded.start;
          DEB("Found from cache: %s", line);
        }
        else 
          DEB("Not found from cache: %s", line);
      }
    }
  }

  /* recreate other blocks from source images */
  list_for_each_entry(source, source_list, list) 
  {
    if (source->encoded.fp == NULL) /* not found from cache */
    {
      if ((source->encoded.fp = tmpfile()) == NULL)
        xperror("tempfile()");

      if (encode_source(source->encoded.fp, source, page))
        xperror(source->fn);

      DEB("Encoded source: %s\n", source->config);

      source->encoded.start = 0;
      source->encoded.length = ftell(source->encoded.fp);
    }
  }
  return 0;
}


int find_ocad_data(FILE *fp, file_block_t *block) {
  char line[LINE_SIZE];

  if(fseek(fp, 0, SEEK_SET))
    return -1;
  
  block->fp = fp;

  while (fgets(line, LINE_SIZE, fp)) {
    if (linecmp(line, "%ASCAD BEGIN DATA") == 0) {
      block->start = ftell(fp);
      while (fgets(line, LINE_SIZE, fp)) {
        if (linecmp(line, "%ASCAD END DATA") == 0) break;
      }
      block->length = ftell(fp) - strlen(line) - block->start;
      return 0;
    }
  }
  block->start = 0;
  block->length = ftell(fp); /*eof*/
  return 0;
}


int copy(FILE *out, FILE *in, size_t size)
{
  char buf[BUF_SIZE];

  while(size > 0)
  {
    int n = (size > BUF_SIZE) ? BUF_SIZE : size;

    if (fread(buf, 1, n, in) != n ||
        fwrite(buf, 1, n, out) != n)
      return -1;
    size -= n;
  }
  return 0;
}

int copy_until(FILE *out, FILE *in, char *end)
{
  char line[LINE_SIZE];

  while (fgets(line, LINE_SIZE, in)) {
    if (end && strcmp(line, end) == 0)
      return 0;
    if (fputs(line, out) < 0)
      return -1;
  }
  return feof(in) != 0;
}

void write_header(FILE *out, char *hash, rect *page)
{
  fprintf(out, "%%!PS-Adobe-3.0 EPSF-3.0\n");
  fprintf(out, "%%%%Creator: %s\n", _version_);
  fprintf(out, "%%%%Pages: 1\n");
  fprintf(out, "%%%%BoundingBox: %.0f %.0f %.0f %.0f\n", page->x1, page->y1, page->x2, page->y2);
  fprintf(out, "%%%%DocumentData: Clean7Bit\n%%%%LanguageLevel: 2\n%%%%EndComments\n");
  fprintf(out, "%%%%Page: 1 1\n");

  fprintf(out, "%%ASCAD HASH %s\n", hash);
  fprintf(out, "\n");
}

int write_encoded_source(FILE *out, char *config, file_block_t *block)
{
  if(fseek(block->fp, block->start, SEEK_SET))
    return -1;

  fprintf(out, "%%ASCAD BEGIN %s\n", config);

  if (copy(out, block->fp, block->length))
    return -1;

  fprintf(out, "%%ASCAD END %s\n", config);

  return 0;
}

int write_ocad_data(FILE *out, file_block_t *block)
{
  fprintf(out, "%%ASCAD BEGIN DATA\n");

  if (block)
  {
    if(fseek(block->fp, block->start, SEEK_SET))
      return -1;

    if (copy(out, block->fp, block->length))
      return -1;
  }

  fprintf(out, "%%ASCAD END DATA\n");

  return 0;
}

void write_trailer(FILE *out)
{
  fprintf(out, "%%ASCAD TRAILER\n");
  fprintf(out, "showpage\n");
  fprintf(out, "%%%%PageTrailer\n");
  fprintf(out, "%%%%Trailer\n");
  fprintf(out, "%%%%EOF\n");
}

void update_bmp(char *bmp_fn, char *eps_fn, map_t *map) {
  char cmd[2048];
  struct stat st;

  // Make bmp from postscript
  sprintf(cmd, "gs -q -dBATCH -dNOPAUSE -g%ldx%ld -r%ld -sDEVICE=bmp16m -sOutputFile=%s %s",
      (long)floor((map->page.x2 - map->page.x1)*map->resolution/72),
      (long)floor((map->page.y2 - map->page.y1)*map->resolution/72),
      (long)floor(map->resolution), bmp_fn, eps_fn);

  if (system(cmd) != 0)
    xerror("system(\"%s\") failed", cmd);

  if (stat(bmp_fn, &st))
    xperror("stat(%s)", bmp_fn);
}

void update_map(map_t *map)
{
  FILE *cache_fp;
  char cache_fn[1024], bmp_fn[1024];
  result_t *result;
  int update_cache = 0;

  sprintf(cache_fn, ".%s.eps", map->name);
  sprintf(bmp_fn, ".%s.bmp", map->name);

  if ((cache_fp = fopen(cache_fn, "r")) == NULL)
  {
    if (errno != ENOENT)
      xperror(cache_fn);

    update_cache = 1;
  }
  else
  {
    char hash[HASH_LEN+1];

    if (read_hash(cache_fp, hash))
      xperror(cache_fn);

    update_cache = (strcmp(hash, map->hash) != 0);
  }

  /* Update .XXXX.eps, .XXXX.bmp and Ocad-files (name of .XXXX.bmp) */
  if (update_cache)
  {
    FILE *fp;
    char *cache_tmp, *bmp_tmp;
    source_t *source;
    ocad_t *ocad;

    fprintf(stderr, "Updating template %s\n", map->name);

    if (find_encodes(cache_fp, &map->source_list, &map->page))
      xperror(cache_fn);

    cache_tmp = strdup2(cache_fn, ".tmp");
    if ((fp = fopen(cache_tmp, "w")) == NULL)
      xperror("fopen(%s)", cache_tmp);

    write_header(fp, map->hash, &map->page);

    list_for_each_entry(source, &map->source_list, list)
      if (write_encoded_source(fp, source->config, &source->encoded))
        xperror(source->fn);

    list_for_each_entry(source, &map->source_list, list)
      if (source->encoded.fp != cache_fp)
        if (fclose(source->encoded.fp))
          xperror("fclose(%s)", source->fn);

    write_trailer(fp);

    if (fclose(fp))
      xperror("fclose(%s)", cache_tmp);

    if (cache_fp && fclose(cache_fp))
      xperror("fclose(%s)", cache_fn);

    bmp_tmp = strdup2(bmp_fn, ".tmp");
    update_bmp(bmp_tmp, cache_tmp, map);

    if (rename(bmp_tmp, bmp_fn))
      xperror("rename(%s, %s)", bmp_tmp, bmp_fn);

    if (rename(cache_tmp, cache_fn))
      xperror("rename(%s, %s)", cache_tmp, cache_fn);
    
    list_for_each_entry(ocad, &map->ocad_list, list) 
    {
      fprintf(stderr, "Updating %s\n", ocad->fn);
      ocd_update_template(ocad->fn, bmp_fn, map->page, map->resolution);
    }

    free(cache_tmp);
    free(bmp_tmp);

    if ((cache_fp = fopen(cache_fn, "r")) == NULL)
      xperror("(re)fopen(%s)", cache_fn);

  }
  else
    fprintf(stderr, "Passed template %s\n", map->name);

  /* Here we always has valid cache_fp */

  /* update RESULTS.eps */
  list_for_each_entry(result, &map->result_list, list) {
    FILE *result_fp;
    char hash[HASH_LEN+1];

    if ((result_fp = fopen(result->fn, "r")) != NULL)
    {
      if (read_hash(result_fp, hash))
        xperror(result->fn);

      if (strcmp(hash, map->hash) != 0)
      {
        file_block_t ocad_data;
        FILE *fp;
        char *result_tmp;

        fprintf(stderr, "Updating result %s\n", result->fn);

        if (find_ocad_data(result_fp, &ocad_data))
          xperror(result->fn);

        result_tmp = strdup2(result->fn, ".tmp");
        if ((fp = fopen(result_tmp, "w")) == NULL)
          xperror(result->fn);

        if(fseek(cache_fp, 0, SEEK_SET))
          xperror("fseek(%s, 0)", cache_fn);

        if (copy_until(fp, cache_fp, "%ASCAD TRAILER\n"))
          xperror("copy_until(%s, %s)", result_tmp, cache_fn);

        if (write_ocad_data(fp, &ocad_data))
          xperror(result->fn);

        write_trailer(fp);

        fclose(result_fp);
        fclose(fp);

        if (rename(result_tmp, result->fn))
          xperror(result_tmp);
      }
      else
        fprintf(stderr, "Passed result %s\n", result->fn);

    } else if (errno != ENOENT) {
      xperror("fopen(%s)", result->fn);

    } else {
      /* skip if the file does not exists */
      fprintf(stderr, "WARNING: %s missing\n", result->fn);
    }
  }

  if (fclose(cache_fp))
    xperror("fclose(%s)", cache_fn);

}

void read_ASconfig(FILE *fh, char *fn, struct list_head *map_list)
{
  map_t *map = NULL;
  char buf[BUF_SIZE];
  int ln=0;

  while (fgets(buf, BUF_SIZE, fh)) {
    char *p0, *p, *line, *c;

    ln++;
    for(line=buf; isspace(*line); line++); // strip leading spaces
    for(c=line + strlen(line) - 1; c>=line && isspace(*c); c--) *c=0; // strip trailing spaces and newline
    if (line[0] == '\0' || line[0] == '#') continue;

    p = p0 = strdup(line);

    if (skip_str(&p, "dospath:")) {

      if (!get_word(&p, &dospath) || *p)
        xerror("AS(%d): Error \"dospath:%s\"\n", ln, p);

    } else if (skip_str(&p, "kartta:")) {

      map = (map_t*)xmalloc(sizeof(map_t));

      if (!get_word(&p, &map->name))
        xerror("AS(%d): Error \"kartta:%s\"\n", ln, p);

      map->config = strdup(line);
      map->page.x1 = map->page.y1 = 0;
      map->page.x2 = 595; map->page.y2 = 842; // A4
      map->resolution = 127;
      INIT_LIST_HEAD(&map->source_list);
      INIT_LIST_HEAD(&map->ocad_list);
      INIT_LIST_HEAD(&map->result_list);

      list_add_tail(&map->list, map_list);

      while(*p) {
        if (skip_str(&p, "koko=")) {
          if (skip_str(&p, "A4R")) map->page.x2 = 842, map->page.y2 = 595;
          else if (skip_str(&p, "A4")) map->page.x2 = 595, map->page.y2 = 842;
          else if (skip_str(&p, "A5R")) map->page.x2 = 595, map->page.y2 = 420;
          else if (skip_str(&p, "A5")) map->page.x2 = 420, map->page.y2 = 595;
          else
            xerror("AS(%d): Unsupporetd (A4,A4R,A5,A5R) \"koko=%s\"", ln, p);
        }
        else if (skip_str(&p, "resoluutio=")) {
          if (!get_num(&p, &map->resolution, 1.0)) 
            xerror("AS(%d): Error \"resoluutio=%s\"\n", ln, p);
        }
        else xerror("AS(%d): Unknown option \"%s\"\n", ln, p);
      }

    }
    else if (skip_str(&p, "pohja:")) {

      char mtime_str[LINE_SIZE];
      source_t *source = xmalloc(sizeof(source_t));

      if (map == NULL)
        xerror("AS(%d): \"kartta:\" undefined", ln);

      if (!get_word(&p, &source->fn))
        xerror("AS(%d): Error \"pohja:%s\"\n", ln, p);

      sprintf(mtime_str, " mtime: %ld", read_mtime(source->fn));
      source->config = strdup2(line, mtime_str);
      source->resolution = 0;
      source->location_type = LOCATION_LOWER_LEFT;
      source->location.x = source->location.y = 0.0;
      source->crop_length = 0; /* 2=rect crop */
      source->scale = 1.00;
      source->rotate = 0.00;
      INIT_LIST_HEAD(&source->col_trans_list);
      INIT_LIST_HEAD(&source->crop_path);
      source->encoded.fp = NULL;

      list_add_tail(&source->list, &map->source_list);

      while(*p) {
        if (skip_str(&p, "rajaa=")) {
          source->crop_box.x1 = source->crop_box.y1 = 99999;
          source->crop_box.x2 = source->crop_box.y2 = -99999;

          do {
            point *pt = xmalloc(sizeof(point));

            if (!get_2num(&p, &pt->x, &pt->y, MM2PT))
              xerror("AS(%d): Error \"rajaa=...%s\"\n", ln, p);

            source->crop_box.x1 = min(source->crop_box.x1, pt->x);
            source->crop_box.y1 = min(source->crop_box.y1, pt->y);
            source->crop_box.x2 = max(source->crop_box.x2, pt->x);
            source->crop_box.y2 = max(source->crop_box.y2, pt->y);

            list_add_tail(&pt->list, &source->crop_path);
            source->crop_length++;

          } while (skip_str(&p, ","));

          if (source->crop_length < 2)
            xerror("AS(%d): Error \"rajaa=...\"\n", ln);
        }
        else if (skip_str(&p, "koillinen=")) {
          source->location_type = LOCATION_UPPER_RIGHT;
          if (!get_2num(&p, &source->location.x, &source->location.y, MM2PT))
            xerror("AS(%d): Error \"koillinen=..%s\"\n", ln, p);
          fit_page(&source->location, &map->page);
        }
        else if (skip_str(&p, "kaakko=")) {
          source->location_type = LOCATION_LOWER_RIGHT;
          if (!get_2num(&p, &source->location.x, &source->location.y, MM2PT))
            xerror("AS(%d): Error \"kaakko=..%s\"\n", ln, p);
          fit_page(&source->location, &map->page);
        }
        else if (skip_str(&p, "lounas=")) {
          source->location_type = LOCATION_LOWER_LEFT;
          if (!get_2num(&p, &source->location.x, &source->location.y, MM2PT))
            xerror("AS(%d): Error \"lounas=..%s\"\n", ln, p);
          fit_page(&source->location, &map->page);
        }
        else if (skip_str(&p, "luode=")) {
          source->location_type = LOCATION_UPPER_LEFT;
          if (!get_2num(&p, &source->location.x, &source->location.y, MM2PT))
            xerror("AS(%d): Error \"luode=..%s\"\n", ln, p);
          fit_page(&source->location, &map->page);
        }
        else if (skip_str(&p, "keskusta=")) {
          source->location_type = LOCATION_CENTER;
          if (!get_2num(&p, &source->location.x, &source->location.y, MM2PT))
            xerror("AS(%d): Error \"keskusta=..%s\"\n", ln, p);
          fit_page(&source->location, &map->page);
        }
        else if (skip_str(&p, "zoom=")) {
          if (!get_num(&p, &source->scale, 0.01))
            xerror("AS(%d): Error \"zoom=%s\"\n", ln, p);
        }
        else if (skip_str(&p, "kierto=")) {
          if (!get_num(&p, &source->rotate, 1.0))
            xerror("AS(%d): Error \"kierto=%s\"\n", ln, p);
        }
        else if (skip_str(&p, "sävy=")) {
          color_transform_t *col_trans = xmalloc(sizeof(color_transform_t));
          col_trans->mode = COL_TRANS_COLORIZE;
          if (skip_str(&p, "ruskea")) {
            col_trans->r = 204; col_trans->g = 102; col_trans->b = 0; /* PMS 138C (?) */
            col_trans->density = 90; /* 10% of original */
          }
          else if (!get_3long(&p, &col_trans->r, &col_trans->g, &col_trans->b) || !!skip_str(&p, ",") ||
                   !get_num(&p, &col_trans->density, 1.0))
            xerror("AS(%d): Error \"sävy=...%s\"\n", ln, p);
          list_add_tail(&col_trans->list, &source->col_trans_list);
        }
        else if (skip_str(&p, "resoluutio=")) {
          if (!get_num(&p, &source->resolution, 1.0))
            xerror("AS(%d): Error \"resoluutio=%s\"\n", ln, p);
        }
        else if (skip_str(&p, "vahvista")) {
          color_transform_t *col_trans = xmalloc(sizeof(color_transform_t));
          col_trans->mode = COL_TRANS_CONTRAST;
          list_add_tail(&col_trans->list, &source->col_trans_list);
        }
        else if (skip_str(&p, "karttapaikka8000")) {
          color_transform_t *col_trans = xmalloc(sizeof(color_transform_t));
          col_trans->mode = COL_TRANS_UNBROWN;
          list_add_tail(&col_trans->list, &source->col_trans_list);
          source->resolution = 254.0; /* pikseli=2m ==> 1:20 000 */
        }
        else if (skip_str(&p, "karttapaikka40000")) {
          source->resolution = 127.0; /* pikseli=10m ==> 1:50 000 */
        }
        else xerror("AS(%d): Unknown option \"%s\"\n", ln, p);
      }


    }
    else if (skip_str(&p, "tulos:")) {

      result_t *result = xmalloc(sizeof(result_t));

      if (map == NULL)
        xerror("AS(%d): \"kartta:\" undefined", ln);

      if(!get_word(&p, &result->fn) || *p)
        xerror("AS(%d): Error \"tulos:%s\"\n", ln, p);
      result->config = strdup(line);

      list_add_tail(&result->list, &map->result_list);

    }
    else if (skip_str(&p, "reitti:")) {

      ocad_t *ocad = xmalloc(sizeof(ocad_t));

      if (map == NULL)
        xerror("AS(%d): \"kartta:\" undefined", ln);

      if (!get_word(&p, &ocad->fn) || *p)
        xerror("AS(%d): Error \"reitti:%s\"\n", ln, p);
      ocad->config = strdup(line);

      list_add_tail(&ocad->list, &map->ocad_list);
    }
    else xerror("AS(%d): Unknown argument \"%s\"\n", ln, p);

    free(p0);
  }
}

int main(int argc, char *argv[]) {
  FILE *fh;
  LIST_HEAD(map_list);
  map_t *map;
  char *fn = "ASCAD";

  errstr = NULL;

  if ((fh=fopen(fn, "r")) == NULL)
    xperror(fn);

  read_ASconfig(fh, fn, &map_list);

  fclose(fh);

  list_for_each_entry(map, &map_list, list) {
    calc_map_hash(map);
    update_map(map);
  }

  return 0;

}


